﻿namespace MyWebLog

open System
open Markdig
open NodaTime

/// <summary>Support functions for domain definition</summary>
[<AutoOpen>]
module private Helpers =

    open Markdown.ColorCode

    /// <summary>Create a new ID (short GUID)</summary>
    /// <returns>A 21-character URL-friendly string representing a GUID</returns>
    /// <remarks>https://www.madskristensen.net/blog/A-shorter-and-URL-friendly-GUID</remarks>
    let newId () =
        Convert.ToBase64String(Guid.NewGuid().ToByteArray()).Replace('/', '_').Replace('+', '-')[..21]

    /// <summary>Pipeline with most extensions enabled</summary>
    let markdownPipeline = MarkdownPipelineBuilder().UseSmartyPants().UseAdvancedExtensions().UseColorCode().Build()


/// <summary>Functions to support NodaTime manipulation</summary>
module Noda =

    /// <summary>The clock to use when getting "now" (will make mutable for testing)</summary>
    let clock: IClock = SystemClock.Instance

    /// <summary>The Unix epoch</summary>
    let epoch = Instant.FromUnixTimeSeconds 0L

    /// <summary>Truncate an instant to remove fractional seconds</summary>
    /// <param name="value">The value from which fractional seconds should be removed</param>
    /// <returns>The <c>Instant</c> value with no fractional seconds</returns>
    let toSecondsPrecision (value: Instant) =
        Instant.FromUnixTimeSeconds(value.ToUnixTimeSeconds())

    /// <summary>The current <c>Instant</c>, with fractional seconds truncated</summary>
    /// <returns>The current <c>Instant</c> with no fractional seconds</returns>
    let now =
        clock.GetCurrentInstant >> toSecondsPrecision

    /// <summary>Convert a date/time to an <c>Instant</c> with whole seconds</summary>
    /// <param name="dt">The date/time to convert</param>
    /// <returns>An <c>Instant</c> with no fractional seconds</returns>
    let fromDateTime (dt: DateTime) =
        Instant.FromDateTimeUtc(DateTime(dt.Ticks, DateTimeKind.Utc)) |> toSecondsPrecision


/// <summary>A user's access level</summary>
[<Struct>]
type AccessLevel =
    /// <summary>The user may create and publish posts and edit the ones they have created</summary>
    | Author
    /// <summary>The user may edit posts they did not create, but may not delete them</summary>
    | Editor
    /// <summary>The user may delete posts and configure web log settings</summary>
    | WebLogAdmin
    /// <summary>The user may manage themes (which affects all web logs for an installation)</summary>
    | Administrator

    /// <summary>Weights applied to each access level</summary>
    static member private Weights =
        [ Author,        10
          Editor,        20
          WebLogAdmin,   30
          Administrator, 40 ]
        |> Map.ofList

    /// <summary>Parse an access level from its string representation</summary>
    /// <param name="level">The string representation to be parsed</param>
    /// <returns>The <c>AccessLevel</c> instance parsed from the string</returns>
    /// <exception cref="InvalidArgumentException">If the string is not valid</exception>
    static member Parse level =
        match level with
        | "Author" -> Author
        | "Editor" -> Editor
        | "WebLogAdmin" -> WebLogAdmin
        | "Administrator" -> Administrator
        | _ -> invalidArg (nameof level) $"{level} is not a valid access level"

    /// <inheritdoc />
    override this.ToString() =
        match this with
        | Author -> "Author"
        | Editor -> "Editor"
        | WebLogAdmin -> "WebLogAdmin"
        | Administrator -> "Administrator"

    /// <summary>Does a given access level allow an action that requires a certain access level?</summary>
    /// <param name="needed">The minimum level of access needed</param>
    /// <returns>True if this level satisfies the given level, false if not</returns>
    member this.HasAccess(needed: AccessLevel) =
        AccessLevel.Weights[needed] <= AccessLevel.Weights[this]


/// <summary>An identifier for a category</summary>
[<Struct>]
type CategoryId =
    | CategoryId of string

    /// <summary>An empty category ID</summary>
    static member Empty = CategoryId ""

    /// <summary>Create a new category ID</summary>
    /// <returns>A new category ID</returns>
    static member Create =
        newId >> CategoryId

    /// <inheritdoc />
    override this.ToString() =
        match this with CategoryId it -> it


/// <summary>An identifier for a comment</summary>
[<Struct>]
type CommentId =
    | CommentId of string

    /// <summary>An empty comment ID</summary>
    static member Empty = CommentId ""

    /// <summary>Create a new comment ID</summary>
    /// <returns>A new commend ID</returns>
    static member Create =
        newId >> CommentId

    /// <inheritdoc />
    override this.ToString() =
        match this with CommentId it -> it


/// <summary>Statuses for post comments</summary>
[<Struct>]
type CommentStatus =
    /// <summary>The comment is approved</summary>
    | Approved
    /// <summary>The comment has yet to be approved</summary>
    | Pending
    /// <summary>The comment was unsolicited and unwelcome</summary>
    | Spam

    /// <summary>Parse a string into a comment status</summary>
    /// <param name="status">The string representation of the status</param>
    /// <returns>The <c>CommentStatus</c> instance parsed from the string</returns>
    /// <exception cref="InvalidArgumentException">If the string is not valid</exception>
    static member Parse status =
        match status with
        | "Approved" -> Approved
        | "Pending" -> Pending
        | "Spam" -> Spam
        | _ -> invalidArg (nameof status) $"{status} is not a valid comment status"

    /// <inheritdoc />
    override this.ToString() =
        match this with Approved -> "Approved" | Pending -> "Pending" | Spam -> "Spam"


/// <summary>Valid values for the iTunes explicit rating</summary>
[<Struct>]
type ExplicitRating =
    | Yes
    | No
    | Clean

    /// <summary>Parse a string into an explicit rating</summary>
    /// <param name="rating">The string representation of the rating</param>
    /// <returns>The <c>ExplicitRating</c> parsed from the string</returns>
    /// <exception cref="InvalidArgumentException">If the string is not valid</exception>
    static member Parse rating =
        match rating with
        | "yes" -> Yes
        | "no" -> No
        | "clean" -> Clean
        | _ -> invalidArg (nameof rating) $"{rating} is not a valid explicit rating"

    /// <inheritdoc />
    override this.ToString() =
        match this with Yes -> "yes" | No -> "no" | Clean -> "clean"


/// <summary>A location (specified by Podcast Index)</summary>
type Location = {
    /// <summary>The name of the location (free-form text)</summary>
    Name: string

    /// <summary>A geographic coordinate string (RFC 5870)</summary>
    Geo: string

    /// <summary>An OpenStreetMap query</summary>
    Osm: string option
}


/// <summary>A chapter in a podcast episode</summary>
type Chapter = {
    /// <summary>The start time for the chapter</summary>
    StartTime: Duration

    /// <summary>The title for this chapter</summary>
    Title: string option

    /// <summary>A URL for an image for this chapter</summary>
    ImageUrl: string option

    /// <summary>A URL with information pertaining to this chapter</summary>
    Url: string option

    /// <summary>Whether this chapter is hidden</summary>
    IsHidden: bool option

    /// <summary>The episode end time for the chapter</summary>
    EndTime: Duration option

    /// <summary>A location that applies to a chapter</summary>
    Location: Location option
} with

    /// <summary>An empty chapter</summary>
    static member Empty =
        { StartTime = Duration.Zero
          Title     = None
          ImageUrl  = None
          Url       = None
          IsHidden  = None
          EndTime   = None
          Location  = None }


open NodaTime.Text

/// <summary>A podcast episode</summary>
type Episode = {
    /// <summary>The URL to the media file for the episode (may be permalink)</summary>
    Media: string

    /// <summary>The length of the media file, in bytes</summary>
    Length: int64

    /// <summary>The duration of the episode</summary>
    Duration: Duration option

    /// <summary>The media type of the file (overrides podcast default if present)</summary>
    MediaType: string option

    /// <summary>The URL to the image file for this episode (overrides podcast image if present, may be permalink)</summary>
    ImageUrl: string option

    /// <summary>A subtitle for this episode</summary>
    Subtitle: string option

    /// <summary>This episode's explicit rating (overrides podcast rating if present)</summary>
    Explicit: ExplicitRating option

    /// <summary>Chapters for this episode</summary>
    Chapters: Chapter list option

    /// <summary>A link to a chapter file</summary>
    ChapterFile: string option

    /// <summary>The MIME type for the chapter file</summary>
    ChapterType: string option

    /// <summary>Whether the chapters have locations that should be displayed as waypoints</summary>
    ChapterWaypoints: bool option

    /// <summary>The URL for the transcript of the episode (may be permalink)</summary>
    TranscriptUrl: string option

    /// <summary>The MIME type of the transcript</summary>
    TranscriptType: string option

    /// <summary>The language in which the transcript is written</summary>
    TranscriptLang: string option

    /// <summary>If true, the transcript will be declared (in the feed) to be a captions file</summary>
    TranscriptCaptions: bool option

    /// <summary>The season number (for serialized podcasts)</summary>
    SeasonNumber: int option

    /// <summary>A description of the season</summary>
    SeasonDescription: string option

    /// <summary>The episode number</summary>
    EpisodeNumber: double option

    /// <summary>A description of the episode</summary>
    EpisodeDescription: string option
} with

    /// <summary>An empty episode</summary>
    static member Empty =
        { Media              = ""
          Length             = 0L
          Duration           = None
          MediaType          = None
          ImageUrl           = None
          Subtitle           = None
          Explicit           = None
          Chapters           = None
          ChapterFile        = None
          ChapterType        = None
          ChapterWaypoints   = None
          TranscriptUrl      = None
          TranscriptType     = None
          TranscriptLang     = None
          TranscriptCaptions = None
          SeasonNumber       = None
          SeasonDescription  = None
          EpisodeNumber      = None
          EpisodeDescription = None }

    /// <summary>Format a duration for an episode</summary>
    /// <returns>A duration formatted in hours, minutes, and seconds</returns>
    member this.FormatDuration() =
        this.Duration |> Option.map (DurationPattern.CreateWithInvariantCulture("H:mm:ss").Format)


/// <summary>Types of markup text</summary>
type MarkupText =
    /// <summary>Markdown text</summary>
    | Markdown of string
    /// <summary>HTML text</summary>
    | Html of string

    /// <summary>Parse a string into a MarkupText instance</summary>
    /// <param name="text">The string to be parsed</param>
    /// <returns>The <c>MarkupText</c> parsed from the string</returns>
    /// <exception cref="InvalidArgumentException">If the string is not valid</exception>
    static member Parse(text: string) =
        match text with
        | _ when text.StartsWith "Markdown: " -> Markdown text[10..]
        | _ when text.StartsWith "HTML: " -> Html text[6..]
        | _ -> invalidArg (nameof text) $"Cannot derive type of text ({text})"

    /// <summary>The source type for the markup text</summary>
    member this.SourceType =
        match this with Markdown _ -> "Markdown" | Html _ -> "HTML"

    /// <summary>The raw text, regardless of type</summary>
    member this.Text =
        match this with Markdown text -> text | Html text -> text

    /// <inheritdoc />
    override this.ToString() =
        $"{this.SourceType}: {this.Text}"

    /// <summary>The HTML representation of the markup text</summary>
    /// <returns>An HTML representation of the markup text</returns>
    member this.AsHtml() =
        match this with Markdown text -> Markdown.ToHtml(text, markdownPipeline) | Html text -> text


/// <summary>An item of metadata</summary>
[<CLIMutable>]
type MetaItem = {
    /// <summary>The name of the metadata value</summary>
    Name: string

    /// <summary>The metadata value</summary>
    Value: string
} with

    /// <summary>An empty metadata item</summary>
    static member Empty =
        { Name = ""; Value = "" }


/// <summary>A revision of a page or post</summary>
[<CLIMutable>]
type Revision = {
    /// <summary>When this revision was saved</summary>
    AsOf: Instant

    /// <summary>The text of the revision</summary>
    Text: MarkupText
} with

    /// <summary>An empty revision</summary>
    static member Empty =
        { AsOf = Noda.epoch; Text = Html "" }


/// <summary>A permanent link</summary>
[<Struct>]
type Permalink =
    | Permalink of string

    /// <summary>An empty permalink</summary>
    static member Empty = Permalink ""

    /// <inheritdoc />
    override this.ToString() =
        match this with Permalink it -> it


/// <summary>An identifier for a page</summary>
[<Struct>]
type PageId =
    | PageId of string

    /// <summary>An empty page ID</summary>
    static member Empty = PageId ""

    /// <summary>Create a new page ID</summary>
    /// <returns>A new page ID</returns>
    static member Create =
        newId >> PageId

    /// <inheritdoc />
    override this.ToString() =
        match this with PageId it -> it


/// <summary>PodcastIndex.org podcast:medium allowed values</summary>
[<Struct>]
type PodcastMedium =
    | Podcast
    | Music
    | Video
    | Film
    | Audiobook
    | Newsletter
    | Blog

    /// <summary>Parse a string into a podcast medium</summary>
    /// <param name="medium">The string to be parsed</param>
    /// <returns>The <c>PodcastMedium</c> parsed from the string</returns>
    /// <exception cref="InvalidArgumentException">If the string is not valid</exception>
    static member Parse medium =
        match medium with
        | "podcast" -> Podcast
        | "music" -> Music
        | "video" -> Video
        | "film" -> Film
        | "audiobook" -> Audiobook
        | "newsletter" -> Newsletter
        | "blog" -> Blog
        | _ -> invalidArg (nameof medium) $"{medium} is not a valid podcast medium"

    /// <inheritdoc />
    override this.ToString() =
        match this with
        | Podcast -> "podcast"
        | Music -> "music"
        | Video -> "video"
        | Film -> "film"
        | Audiobook -> "audiobook"
        | Newsletter -> "newsletter"
        | Blog -> "blog"


/// <summary>Statuses for posts</summary>
[<Struct>]
type PostStatus =
    /// <summary>The post should not be publicly available</summary>
    | Draft
    /// <summary>The post is publicly viewable</summary>
    | Published

    /// <summary>Parse a string into a post status</summary>
    /// <param name="status">The string to be parsed</param>
    /// <returns>The <c>PostStatus</c> parsed from the string</returns>
    /// <exception cref="InvalidArgumentException">If the string is not valid</exception>
    static member Parse status =
        match status with
        | "Draft" -> Draft
        | "Published" -> Published
        | _ -> invalidArg (nameof status) $"{status} is not a valid post status"

    /// <summary>The string representation of this post status</summary>
    override this.ToString() =
        match this with Draft -> "Draft" | Published -> "Published"


/// <summary>An identifier for a post</summary>
[<Struct>]
type PostId =
    | PostId of string

    /// <summary>An empty post ID</summary>
    static member Empty = PostId ""

    /// <summary>Create a new post ID</summary>
    /// <returns>A new post ID</returns>
    static member Create =
        newId >> PostId

    /// <inheritdoc />
    override this.ToString() =
        match this with PostId it -> it


/// <summary>A redirection for a previously valid URL</summary>
[<CLIMutable>]
type RedirectRule = {
    /// <summary>The From string or pattern</summary>
    From: string

    /// <summary>The To string or pattern</summary>
    To: string

    /// <summary>Whether to use regular expressions on this rule</summary>
    IsRegex: bool
} with

    /// <summary>An empty redirect rule</summary>
    static member Empty =
        { From = ""; To = ""; IsRegex = false }


/// <summary>An identifier for a custom feed</summary>
[<Struct>]
type CustomFeedId =
    | CustomFeedId of string

    /// <summary>An empty custom feed ID</summary>
    static member Empty = CustomFeedId ""

    /// <summary>Create a new custom feed ID</summary>
    /// <returns>A new custom feed ID</returns>
    static member Create =
        newId >> CustomFeedId

    /// <inheritdoc />
    override this.ToString() =
        match this with CustomFeedId it -> it


/// <summary>The source for a custom feed</summary>
type CustomFeedSource =
    /// <summary>A feed based on a particular category</summary>
    | Category of CategoryId
    /// <summary>A feed based on a particular tag</summary>
    | Tag of string

    /// <summary>Parse a feed source from its string version</summary>
    /// <param name="source">The string to be parsed</param>
    /// <returns>The <c>CustomFeedSource</c> parsed from the string</returns>
    /// <exception cref="InvalidArgumentException">If the string is not valid</exception>
    static member Parse(source: string) =
        let value (it : string) = it.Split(":").[1]
        match source with
        | _ when source.StartsWith "category:" -> (value >> CategoryId >> Category) source
        | _ when source.StartsWith "tag:"      -> (value >> Tag) source
        | _ -> invalidArg (nameof source) $"{source} is not a valid feed source"

    /// <inheritdoc />
    override this.ToString() =
        match this with | Category (CategoryId catId) -> $"category:{catId}" | Tag tag -> $"tag:{tag}"


/// <summary>Options for a feed that describes a podcast</summary>
[<CLIMutable; NoComparison; NoEquality>]
type PodcastOptions = {
    /// <summary>The title of the podcast</summary>
    Title: string

    /// <summary>A subtitle for the podcast</summary>
    Subtitle: string option

    /// <summary>The number of items in the podcast feed</summary>
    ItemsInFeed: int

    /// <summary>A summary of the podcast (iTunes field)</summary>
    Summary: string

    /// <summary>The display name of the podcast author (iTunes field)</summary>
    DisplayedAuthor: string

    /// <summary>The e-mail address of the user who registered the podcast at iTunes</summary>
    Email: string

    /// <summary>The link to the image for the podcast</summary>
    ImageUrl: Permalink

    /// <summary>The category from Apple Podcasts (iTunes) under which this podcast is categorized</summary>
    AppleCategory: string

    /// <summary>
    /// A further refinement of the categorization of this podcast (Apple Podcasts/iTunes field / values)
    /// </summary>
    AppleSubcategory: string option

    /// <summary>The explictness rating (iTunes field)</summary>
    Explicit: ExplicitRating

    /// <summary>The default media type for files in this podcast</summary>
    DefaultMediaType: string option

    /// <summary>
    /// The base URL for relative URL media files for this podcast (optional; defaults to web log base)
    /// </summary>
    MediaBaseUrl: string option

    /// <summary>A GUID for this podcast</summary>
    PodcastGuid: Guid option

    /// <summary>A URL at which information on supporting the podcast may be found (supports permalinks)</summary>
    FundingUrl: string option

    /// <summary>The text to be displayed in the funding item within the feed</summary>
    FundingText: string option

    /// <summary>The medium (what the podcast IS, not what it is ABOUT)</summary>
    Medium: PodcastMedium option
} with

    /// <summary>A default set of podcast options</summary>
    static member Empty =
        { Title            = ""
          Subtitle         = None
          ItemsInFeed      = 0
          Summary          = ""
          DisplayedAuthor  = ""
          Email            = ""
          ImageUrl         = Permalink.Empty
          AppleCategory    = ""
          AppleSubcategory = None
          Explicit         = No
          DefaultMediaType = None
          MediaBaseUrl     = None
          PodcastGuid      = None
          FundingUrl       = None
          FundingText      = None
          Medium           = None }


/// <summary>A custom feed</summary>
[<CLIMutable; NoComparison; NoEquality>]
type CustomFeed = {
    /// <summary>The ID of the custom feed</summary>
    Id: CustomFeedId

    /// <summary>The source for the custom feed</summary>
    Source: CustomFeedSource

    /// <summary>The path for the custom feed</summary>
    Path: Permalink

    /// <summary>Podcast options, if the feed defines a podcast</summary>
    Podcast: PodcastOptions option
} with

    /// <summary>An empty custom feed</summary>
    static member Empty =
        { Id      = CustomFeedId.Empty
          Source  = Category CategoryId.Empty
          Path    = Permalink.Empty
          Podcast = None }


/// <summary>Really Simple Syndication (RSS) options for this web log</summary>
[<CLIMutable; NoComparison; NoEquality>]
type RssOptions = {
    /// <summary>Whether the site feed of posts is enabled</summary>
    IsFeedEnabled: bool

    /// <summary>The name of the file generated for the site feed</summary>
    FeedName: string

    /// <summary>Override the "posts per page" setting for the site feed</summary>
    ItemsInFeed: int option

    /// <summary>Whether feeds are enabled for all categories</summary>
    IsCategoryEnabled: bool

    /// <summary>Whether feeds are enabled for all tags</summary>
    IsTagEnabled: bool

    /// <summary>A copyright string to be placed in all feeds</summary>
    Copyright: string option

    /// <summary>Custom feeds for this web log</summary>
    CustomFeeds: CustomFeed list
} with

    /// <summary>An empty set of RSS options</summary>
    static member Empty =
        { IsFeedEnabled     = true
          FeedName          = "feed.xml"
          ItemsInFeed       = None
          IsCategoryEnabled = true
          IsTagEnabled      = true
          Copyright         = None
          CustomFeeds       = [] }


/// <summary>An identifier for a tag mapping</summary>
[<Struct>]
type TagMapId =
    | TagMapId of string

    /// <summary>An empty tag mapping ID</summary>
    static member Empty = TagMapId ""

    /// <summary>Create a new tag mapping ID</summary>
    /// <returns>A new tag mapping ID</returns>
    static member Create =
        newId >> TagMapId

    /// <inheritdoc />
    override this.ToString() =
        match this with TagMapId it -> it


/// <summary>An identifier for a theme (represents its path)</summary>
[<Struct>]
type ThemeId =
    | ThemeId of string

    /// <summary>An empty theme ID</summary>
    static member Empty = ThemeId ""

    /// <inheritdoc />
    override this.ToString() =
        match this with ThemeId it -> it


/// <summary>An identifier for a theme asset</summary>
[<Struct>]
type ThemeAssetId =
    | ThemeAssetId of ThemeId * string

    /// <summary>An empty theme asset ID</summary>
    static member Empty = ThemeAssetId(ThemeId.Empty, "")

    /// <summary>Convert a string into a theme asset ID</summary>
    /// <param name="it">The string to be parsed</param>
    /// <returns>The <c>ThemeAssetId</c> parsed from the string</returns>
    /// <exception cref="InvalidArgumentException">If the string is not valid</exception>
    static member Parse(it : string) =
        let themeIdx = it.IndexOf "/"
        if themeIdx < 0 then
            invalidArg "id" $"Invalid format; expected [theme_id]/[asset_id] (received {it})"
        else
            ThemeAssetId(ThemeId it[..(themeIdx - 1)], it[(themeIdx + 1)..])

    /// <inheritdoc />
    override this.ToString() =
        match this with ThemeAssetId (ThemeId theme, asset) -> $"{theme}/{asset}"


/// <summary>A template for a theme</summary>
[<CLIMutable; NoComparison; NoEquality>]
type ThemeTemplate = {
    /// <summary>The name of the template</summary>
    Name: string

    /// <summary>The text of the template</summary>
    Text: string
} with

    /// <summary>An empty theme template</summary>
    static member Empty =
        { Name = ""; Text = "" }


/// <summary>Where uploads should be placed</summary>
[<Struct>]
type UploadDestination =
    | Database
    | Disk

    /// <summary>Parse an upload destination from its string representation</summary>
    /// <param name="destination">The string to be parsed</param>
    /// <returns>The <c>UploadDestination</c> parsed from the string</returns>
    /// <exception cref="InvalidArgumentException">If the string is not valid</exception>
    static member Parse destination =
        match destination with
        | "Database" -> Database
        | "Disk" -> Disk
        | _ -> invalidArg (nameof destination) $"{destination} is not a valid upload destination"

    /// <inheritdoc />
    override this.ToString() =
        match this with Database -> "Database" | Disk -> "Disk"


/// <summary>An identifier for an upload</summary>
[<Struct>]
type UploadId =
    | UploadId of string

    /// <summary>An empty upload ID</summary>
    static member Empty = UploadId ""

    /// <summary>Create a new upload ID</summary>
    /// <returns>A new upload ID</returns>
    static member Create =
        newId >> UploadId

    /// <inheritdoc />
    override this.ToString() =
        match this with UploadId it -> it


/// <summary>An identifier for a web log</summary>
[<Struct>]
type WebLogId =
    | WebLogId of string

    /// <summary>An empty web log ID</summary>
    static member Empty = WebLogId ""

    /// <summary>Create a new web log ID</summary>
    /// <returns>A new web log ID</returns>
    static member Create =
        newId >> WebLogId

    /// <inheritdoc />
    override this.ToString() =
        match this with WebLogId it -> it


/// <summary>An identifier for a web log user</summary>
[<Struct>]
type WebLogUserId =
    | WebLogUserId of string

    /// <summary>An empty web log user ID</summary>
    static member Empty = WebLogUserId ""

    /// <summary>Create a new web log user ID</summary>
    /// <returns>A new web log user ID</returns>
    static member Create =
        newId >> WebLogUserId

    /// <inheritdoc />
    override this.ToString() =
        match this with WebLogUserId it -> it
